/*
 * process, Distribution repositories and basic setup for Emmanuel Bigeon projects
 * Copyright (C) 2014-2018 E. Bigeon
 * mailto:emmanuel@bigeon.fr
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * 
 */
package net.bigeon.gclc.process.io;

/*-
 * #%L
 * process
 * %%
 * Copyright (C) 2014 - 2018 Bigeon
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.bigeon.gclc.manager.ConsoleInput;
import net.bigeon.gclc.tools.StringProvider;

/** @author Emmanuel Bigeon */
public final class ConnectingConsoleInput implements ConsoleInput {

    private static final Logger LOGGER = Logger
            .getLogger(ConnectingConsoleInput.class.getName());
    private boolean             close  = false;
    private StringProvider      prompt;
    private boolean             prompting;
    private final Object        promptLock     = new Object();
    private final Object        connectionLock = new Object();
    private ConsoleInput        connected;
    private boolean             disconnection;

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#close() */
    @Override
    public void close() throws IOException {
        close = true;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#getPrompt() */
    @Override
    public StringProvider getPrompt() {
        return prompt;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#interruptPrompt() */
    @Override
    public void interruptPrompt() {
        synchronized (promptLock) {
            prompting = false;
            if (connected != null) {
                connected.interruptPrompt();
            }
            promptLock.notifyAll();
        }
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#isClosed() */
    @Override
    public boolean isClosed() {
        return close;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt() */
    @Override
    public String prompt() throws IOException {
        return prompt(prompt.apply());
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#setPrompt(fr.bigeon.gclc.tools.
     * StringProvider) */
    @Override
    public void setPrompt(StringProvider string) {
        prompt = string;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(long) */
    @Override
    public String prompt(long timeout) throws IOException {
        return prompt(prompt.apply(), timeout);
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(java.lang.String) */
    @Override
    public String prompt(String message) throws IOException {
        synchronized (promptLock) {
            prompting = true;
        }
        while (prompting) {
            synchronized (promptLock) {
                if (connected == null) {
                    try {
                        promptLock.wait();
                    } catch (final InterruptedException e) {
                        LOGGER.log(Level.WARNING, "Inerruption of console thread", e);
                        Thread.currentThread().interrupt();
                    }
                } else {
                    final String res = connected.prompt(message);
                    synchronized (connectionLock) {
                        if (disconnection) {
                            disconnection = false;
                        } else if (prompting) {
                            return res;
                        }
                    }
                }
            }
        }
        return null;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(java.lang.String, long) */
    @Override
    public String prompt(String message, long timeout) throws IOException {
        if (timeout <= 0) {
            return prompt(message);
        }
        final long end = System.currentTimeMillis() + timeout;
        synchronized (promptLock) {
            prompting = true;
        }
        while (prompting) {
            synchronized (promptLock) {
                if (connected == null) {
                    try {
                        promptLock.wait();
                    } catch (final InterruptedException e) {
                        LOGGER.log(Level.WARNING, "Inerruption of console thread", e);
                        Thread.currentThread().interrupt();
                    }
                } else {
                    final String res = connected.prompt(message,
                            end - System.currentTimeMillis());
                    synchronized (connectionLock) {
                        if (disconnection) {
                            disconnection = false;
                        } else if (prompting) {
                            return res;
                        }
                    }
                }
            }
        }
        return null;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#setPrompt(java.lang.String) */
    @Override
    public void setPrompt(final String prompt) {
        this.prompt = new StringProvider() {
            @Override
            public String apply() {
                return prompt;
            }
        };
    }

    public void connect(ConsoleInput input) {
        disconnect();
        synchronized (promptLock) {
            connected = input;
            promptLock.notifyAll();
        }
    }

    public void disconnect() {
        synchronized (connectionLock) {
            if (connected != null) {
                disconnection = true;
                synchronized (promptLock) {
                    connected.interruptPrompt();
                }
                connected = null;
            }
        }
    }
}
